"""From Co-pilot"""

import unittest

from datetime import datetime, timedelta

import particle_tracking_manager as ptm


# from ptm.the_manager import ParticleTrackingManager


class TestParticleTrackingManager(unittest.TestCase):
    def setUp(self):
        self.ptm = ptm.ParticleTrackingManager(model="opendrift")

    def test_init(self):
        self.assertEqual(self.ptm.model, "opendrift")
        self.assertIsNone(self.ptm.lon)
        self.assertIsNone(self.ptm.lat)
        self.assertIsNone(self.ptm.start_time)
        self.assertIsNone(self.ptm.ocean_model)
        self.assertIsNone(self.ptm.surface_only)
        self.assertEqual(self.ptm.log, "low")

    def test_set_lon(self):
        self.ptm.lon = 50
        self.assertEqual(self.ptm.lon, 50)

    def test_set_lat(self):
        self.ptm.lat = 30
        self.assertEqual(self.ptm.lat, 30)

    def test_set_start_time(self):
        start_time = datetime.now()
        self.ptm.start_time = start_time
        self.assertEqual(self.ptm.start_time, start_time)

    def test_set_ocean_model(self):
        self.ptm.ocean_model = "CIOFS"
        self.assertEqual(self.ptm.ocean_model, "CIOFS")

    def test_set_surface_only(self):
        self.ptm.surface_only = True
        self.assertEqual(self.ptm.surface_only, True)
        self.assertEqual(self.ptm.do3D, False)
        self.assertEqual(self.ptm.z, 0)
        self.assertEqual(self.ptm.vertical_mixing, False)

    def test_set_do3D(self):
        self.ptm.do3D = True
        self.assertEqual(self.ptm.do3D, True)
        self.assertEqual(self.ptm.vertical_mixing, False)

    def test_set_seed_seafloor(self):
        self.ptm.seed_seafloor = True
        self.assertEqual(self.ptm.seed_seafloor, True)
        self.assertEqual(self.ptm.z, None)

    def test_set_time_step(self):
        self.ptm.time_step = 3600
        self.assertEqual(self.ptm.time_step, 3600)

    def test_set_time_step_output(self):
        self.ptm.time_step_output = 3600
        self.assertEqual(self.ptm.time_step_output, 3600)

    def test_set_steps(self):
        self.ptm.steps = 10
        self.assertEqual(self.ptm.steps, 10)

    def test_set_duration(self):
        duration = timedelta(hours=48)
        self.ptm.duration = duration
        self.assertEqual(self.ptm.duration, duration)

    def test_set_end_time(self):
        end_time = datetime.now() + timedelta(hours=48)
        self.ptm.end_time = end_time
        self.assertEqual(self.ptm.end_time, end_time)


if __name__ == "__main__":
    unittest.main()



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import logging

from bisect import bisect_left, bisect_right
from datetime import datetime


logger = logging.getLogger(__name__)

import numpy as np
import xarray as xr

from netCDF4 import num2date
from opendrift.readers.basereader import BaseReader, StructuredReader, vector_pairs_xy
from opendrift.readers.roppy import depth


class Reader(BaseReader, StructuredReader):
    """Reader for ROMS models."""

    def __init__(
        self,
        filename=None,
        name=None,
        gridfile=None,
        standard_name_mapping={},
        kwargs_grid_merge=None,
        kwargs_xarray=None,
    ):
        """Initialize ROMS model reader."""

        if filename is None:
            raise ValueError("Need filename as argument to constructor")

        # Map ROMS variable names to CF standard_name
        self.ROMS_variable_mapping = {
            # Removing (temoprarily) land_binary_mask from ROMS-variables,
            # as this leads to trouble with linearNDFast interpolation
            # 'mask_rho': 'land_binary_mask',
            # 'mask_psi': 'land_binary_mask',
            "wetdry_mask_rho": "land_binary_mask",
            "h": "sea_floor_depth_below_sea_level",
            "zeta": "sea_surface_height",
            "u": "x_sea_water_velocity",
            "v": "y_sea_water_velocity",
            "u_eastward": "x_sea_water_velocity",  # for NWGOA but not CIOFS
            "v_northward": "y_sea_water_velocity",
            "w": "upward_sea_water_velocity",
            "temp": "sea_water_temperature",
            "salt": "sea_water_salinity",
            "uice": "sea_ice_x_velocity",
            "vice": "sea_ice_y_velocity",
            "aice": "sea_ice_area_fraction",
            "hice": "sea_ice_thickness",
            # 'gls': 'turbulent_generic_length_scale',
            # 'tke': 'turbulent_kinetic_energy',
            # 'AKs': 'ocean_vertical_diffusivity',
            # 'sustr': 'surface_downward_x_stress',
            # 'svstr': 'surface_downward_y_stress',
            # 'tair': 'air_temperature',
            "wspd": "wind_speed",
            "uwnd": "x_wind",
            "vwnd": "y_wind",
            "uwind": "x_wind",
            "vwind": "y_wind",
            # CIOFS but these are east/north too
            # will be rotated
            "Uwind": "x_wind",
            "Vwind": "y_wind",
            # NWGOA, there are east/north oriented and will not be rotated
            # because "east" "north" in variable names
            "Uwind_eastward": "x_wind",
            "Vwind_northward": "y_wind",
        }

        # Add user provided variable mappings
        self.ROMS_variable_mapping.update(standard_name_mapping)

        # z-levels to which sigma-layers may be interpolated
        self.zlevels = np.array(
            [
                0,
                -0.5,
                -1,
                -3,
                -5,
                -10,
                -25,
                -50,
                -75,
                -100,
                -150,
                -200,
                -250,
                -300,
                -400,
                -500,
                -600,
                -700,
                -800,
                -900,
                -1000,
                -1500,
                -2000,
                -2500,
                -3000,
                -3500,
                -4000,
                -4500,
                -5000,
                -5500,
                -6000,
                -6500,
                -7000,
                -7500,
                -8000,
            ]
        )

        gls_param = ["gls_cmu0", "gls_p", "gls_m", "gls_n"]

        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        kwargs_xarray = kwargs_xarray or {}
        try:
            # Open file, check that everything is ok
            logger.info("Opening dataset: " + filestr)
            if ("*" in filestr) or ("?" in filestr) or ("[" in filestr):
                logger.info("Opening files with MFDataset")

                def drop_non_essential_vars_pop(ds):
                    """Sneaky function"""
                    dropvars = [
                        v
                        for v in ds.variables
                        if v
                        not in list(self.ROMS_variable_mapping.keys())
                        + gls_param
                        + [
                            "ocean_time",
                            "time",
                            "bulk_time",
                            "s_rho",
                            "Cs_r",
                            "hc",
                            "angle",
                            "Vtransform",
                        ]
                        and v[0:3] not in ["lon", "lat", "mas"]
                    ]
                    logger.debug("Dropping variables: %s" % dropvars)
                    ds = ds.drop_vars(dropvars)
                    return ds

                self.Dataset = xr.open_mfdataset(
                    filename,
                    chunks={"ocean_time": 1},
                    compat="override",
                    decode_times=False,
                    preprocess=drop_non_essential_vars_pop,
                    data_vars="minimal",
                    coords="minimal",
                )
            else:
                logger.info("Opening file with Dataset")
                self.Dataset = xr.open_dataset(
                    filename, decode_times=False, **kwargs_xarray
                )
        except Exception as e:
            raise ValueError(e)

        kwargs_grid_merge = kwargs_grid_merge or {}
        if gridfile is not None:  # Merging gridfile dataset with main dataset
            gf = xr.open_dataset(gridfile)
            self.Dataset = xr.merge([self.Dataset, gf], **kwargs_grid_merge)

        if "Vtransform" in self.Dataset.variables:
            self.Vtransform = self.Dataset.variables["Vtransform"].data  # scalar
        else:
            logger.warning("Vtransform not found, using 1")
            self.Vtransform = 1
        self.Vtransform = np.asarray(self.Vtransform)

        if "s_rho" not in self.Dataset.variables:
            dimensions = 2
        else:
            dimensions = 3

        if dimensions == 3:
            # Read sigma-coordinate values
            try:
                self.sigma = self.Dataset.variables["s_rho"][:]
                self.sigma = self.sigma.data  # Extract data
            except:
                num_sigma = len(self.Dataset.dimensions["s_rho"])
                logger.warning(
                    "s_rho not available in dataset, constructing from"
                    " number of layers (%s)." % num_sigma
                )
                self.sigma = (np.arange(num_sigma) + 0.5 - num_sigma) / num_sigma

            # Read sigma-coordinate transform parameters
            try:
                self.Dataset.variables["Cs_r"].set_auto_mask(False)
            except:
                pass
            self.Cs_r = self.Dataset.variables["Cs_r"][:]
            try:
                self.hc = self.Dataset.variables["hc"][:]
            except:
                self.hc = self.Dataset.variables["hc"].data  # scalar

            self.num_layers = len(self.sigma)
        else:
            self.num_layers = 1
            self.ROMS_variable_mapping["ubar"] = "x_sea_water_velocity"
            self.ROMS_variable_mapping["vbar"] = "y_sea_water_velocity"
            del self.ROMS_variable_mapping["u"]
            del self.ROMS_variable_mapping["v"]

        if "lat_rho" in self.Dataset.variables:
            # Horizontal oordinates and directions
            self.lat = self.Dataset.variables["lat_rho"][:]
            self.lon = self.Dataset.variables["lon_rho"][:]
            self.lon = self.lon.data  # Extract, could be avoided downstream
            self.lat = self.lat.data
            if self.lat.ndim == 1:
                self.lon, self.lat = np.meshgrid(self.lon, self.lat)
                self.angle_xi_east = 0
        else:
            raise ValueError(
                filename + " does not contain lon/lat "
                'arrays, please supply a grid-file: "gridfile=<grid_file>"'
            )

        # For NWGOA, need to calculate wetdry mask from a variable
        if "wetdry_mask_rho" not in self.Dataset and "zeta" in self.Dataset.data_vars:
            self.Dataset["wetdry_mask_rho"] = (~self.Dataset.zeta.isnull()).astype(int)

        for var in list(self.ROMS_variable_mapping):  # Remove unused variables
            if var not in self.Dataset.variables:
                del self.ROMS_variable_mapping[var]

        try:  # Check for GLS parameters (diffusivity)
            self.gls_parameters = {}
            for gls_par in gls_param:
                self.gls_parameters[gls_par] = self.Dataset.variables[gls_par][()]
            logger.info("Read GLS parameters from file.")
        except Exception as e:
            logger.info(e)
            logger.info("Did not find complete set of GLS parameters")

        # Get time coverage
        ocean_time = None
        for tv in ["ocean_time", "time", "bulk_time"]:
            if tv in self.Dataset.variables:
                ocean_time = self.Dataset.variables[tv]
        if ocean_time is None:
            raise ValueError("Time variable not found in ROMS file")
        time_units = ocean_time.attrs["units"]
        if time_units == "second":
            logger.info(
                "Ocean time given as seconds relative to start "
                "Setting artificial start time of 1 Jan 2000."
            )
            time_units = "seconds since 2000-01-01 00:00:00"
        if time_units == "days":
            logger.info(
                "Ocean time given as days relative to start "
                "Setting artificial start time of 1 Jan 2000."
            )
            time_units = "days since 2000-01-01 00:00:00"
        self.times = num2date(ocean_time[:], time_units)
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        if len(self.times) > 1:
            self.time_step = self.times[1] - self.times[0]
        else:
            self.time_step = None

        self.name = "roms native"

        self.precalculate_s2z_coefficients = True

        # Find all variables having standard_name
        self.variables = []
        for var_name in list(self.Dataset.variables):
            var = self.Dataset.variables[var_name]
            if (
                "standard_name" in var.attrs
                and var_name not in self.ROMS_variable_mapping.keys()
            ):
                self.ROMS_variable_mapping[var_name] = var.attrs["standard_name"]
            if var_name in self.ROMS_variable_mapping.keys():
                self.variables.append(self.ROMS_variable_mapping[var_name])

        # A bit hackish solution:
        # If variable names or their standard_name contain "east" or "north",
        # these should not be rotated from xi-direction to east-direction
        # Most or no models will actually use "east/north" variables though
        self.do_not_rotate = []
        for var, stdname in self.ROMS_variable_mapping.items():
            if (
                "east" in var.lower()
                or "east" in stdname.lower()
                or "north" in var.lower()
                or "north" in stdname.lower()
            ):
                self.do_not_rotate.append(stdname)
        if len(self.do_not_rotate) > 0:
            logger.debug(
                "The following ROMS vectors are considered east-north, and will not be rotated %s"
                % self.do_not_rotate
            )

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requested_variables, time=None, x=None, y=None, z=None):
        """Get variables from model output files."""

        start_time = datetime.now()
        requested_variables, time, x, y, z, outside = self.check_arguments(
            requested_variables, time, x, y, z
        )

        # if 'land_binary_mask' in requested_variables and not hasattr(self, 'land_binary_mask'):
        #     self.land_binary_mask = 1 - self.Dataset.variables['wetdry_mask_rho']
        #     # Read landmask for whole domain, for later re-use
        #     # self.land_binary_mask = 1 - self.Dataset.variables['mask_rho'][:]

        if "sea_floor_depth_below_sea_level" in requested_variables and not hasattr(
            self, "sea_floor_depth_below_sea_level"
        ):
            self.sea_floor_depth_below_sea_level = self.Dataset.variables["h"][:]

        # If one vector component is requested, but not the other
        # we must add the other for correct rotation
        for vector_pair in vector_pairs_xy:
            if (
                vector_pair[0] in requested_variables
                and vector_pair[1] not in requested_variables
            ):
                requested_variables.extend([vector_pair[1]])
            if (
                vector_pair[1] in requested_variables
                and vector_pair[0] not in requested_variables
            ):
                requested_variables.extend([vector_pair[0]])

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = self.nearest_time(time)

        variables = {}

        if z is None:
            z = np.atleast_1d(0)

        # Find horizontal indices corresponding to requested x and y
        if hasattr(self, "clipped"):
            clipped = self.clipped
        else:
            clipped = 0
        indx = np.floor((x - self.xmin) / self.delta_x).astype(int) + clipped
        indy = np.floor((y - self.ymin) / self.delta_y).astype(int) + clipped
        indx[outside] = 0  # To be masked later
        indy[outside] = 0
        indx_el = indx.copy()
        indy_el = indy.copy()
        # Adding buffer, to cover also future positions of elements
        buffer = self.buffer
        # Avoiding the last pixel in each dimension, since there are
        # several grids which are shifted (rho, u, v, psi)
        indx = np.arange(
            np.max([0, indx.min() - buffer]),
            np.min([indx.max() + buffer, self.lon.shape[1] - 1]),
        )
        indy = np.arange(
            np.max([0, indy.min() - buffer]),
            np.min([indy.max() + buffer, self.lon.shape[0] - 1]),
        )

        # Find depth levels covering all elements
        if z.min() == 0 or not hasattr(self, "hc"):
            indz = self.num_layers - 1  # surface layer
            variables["z"] = 0

        else:
            # Find the range of indices covering given z-values
            if not hasattr(self, "sea_floor_depth_below_sea_level"):
                logger.debug("Reading sea floor depth...")
                self.sea_floor_depth_below_sea_level = self.Dataset.variables["h"][:]

            if not hasattr(self, "z_rho_tot"):
                Htot = self.sea_floor_depth_below_sea_level
                self.z_rho_tot = depth.sdepth(
                    Htot, self.hc, self.Cs_r, Vtransform=self.Vtransform
                )

            H = self.sea_floor_depth_below_sea_level[indy, indx]
            z_rho = depth.sdepth(H, self.hc, self.Cs_r, Vtransform=self.Vtransform)
            # Element indices must be relative to extracted subset
            indx_el = np.clip(indx_el - indx.min(), 0, z_rho.shape[2] - 1)
            indy_el = np.clip(indy_el - indy.min(), 0, z_rho.shape[1] - 1)

            # Loop to find the layers covering the requested z-values
            indz_min = 0
            indz_max = self.num_layers
            for i in range(self.num_layers):
                if np.min(z - z_rho[i, indy_el, indx_el]) > 0:
                    indz_min = i
                if np.max(z - z_rho[i, indy_el, indx_el]) > 0:
                    indz_max = i
            indz = range(
                np.maximum(0, indz_min - self.verticalbuffer),
                np.minimum(self.num_layers, indz_max + 1 + self.verticalbuffer),
            )
            z_rho = z_rho[indz, :, :]
            # Determine the z-levels to which to interpolate
            zi1 = np.maximum(
                0, bisect_left(-np.array(self.zlevels), -z.max()) - self.verticalbuffer
            )
            zi2 = np.minimum(
                len(self.zlevels),
                bisect_right(-np.array(self.zlevels), -z.min()) + self.verticalbuffer,
            )
            variables["z"] = np.array(self.zlevels[zi1:zi2])

        # read_masks = {}  # To store maskes for various grids
        mask_values = {}
        for par in requested_variables:
            varname = [
                name for name, cf in self.ROMS_variable_mapping.items() if cf == par
            ]
            var = self.Dataset.variables[varname[0]]

            if par == "land_binary_mask":
                if var.ndim == 2:
                    variables[par] = self.Dataset["wetdry_mask_rho"][indy, indx]
                elif var.ndim == 3:
                    variables[par] = self.Dataset["wetdry_mask_rho"][
                        indxTime, indy, indx
                    ]
                variables[par] = 1 - variables[par]
            elif par == "sea_floor_depth_below_sea_level":
                variables[par] = self.sea_floor_depth_below_sea_level[indy, indx]
            elif var.ndim == 2:
                variables[par] = var[indy, indx]
            elif var.ndim == 3:
                variables[par] = var[indxTime, indy, indx]
            elif var.ndim == 4:
                variables[par] = var[indxTime, indz, indy, indx]
            else:
                raise Exception(
                    "Wrong dimension of variable: " + self.variable_mapping[par]
                )

            variables[par] = np.asarray(variables[par])  # If Xarray
            start = datetime.now()

            if par not in mask_values:
                indxgrid = indx
                indygrid = indy
                if par in [
                    "x_sea_water_velocity",
                    "sea_water_x_velocity",
                    "eastward_sea_water_velocity",
                ]:
                    if not hasattr(self, "mask_u"):
                        if "mask_u" in self.Dataset.variables:
                            self.mask_u = self.Dataset.variables["mask_u"][:]
                        elif "mask_rho" in self.Dataset.variables:
                            self.mask_u = self.Dataset.variables["mask_rho"][:]
                        else:
                            continue
                    mask = self.mask_u[indygrid, indxgrid]
                elif par in [
                    "y_sea_water_velocity",
                    "sea_water_y_velocity",
                    "northward_sea_water_velocity",
                ]:
                    if not hasattr(self, "mask_v"):
                        if "mask_v" in self.Dataset.variables:
                            self.mask_v = self.Dataset.variables["mask_v"][:]
                        elif "mask_rho" in self.Dataset.variables:
                            self.mask_v = self.Dataset.variables["mask_rho"][:]
                        else:
                            continue
                    mask = self.mask_v[indygrid, indxgrid]
                else:
                    if not hasattr(self, "land_binary_mask"):
                        # For ROMS-Agrif this must perhaps be mask_psi?
                        if "mask_rho" in self.Dataset.variables:
                            self.land_binary_mask = (
                                1 - self.Dataset.variables["mask_rho"][:]
                            )
                        elif "mask_psi" in self.Dataset.variables:
                            self.land_binary_mask = (
                                1 - self.Dataset.variables["mask_psi"][:]
                            )
                    mask = 1 - self.land_binary_mask[indygrid, indxgrid]
                mask = np.asarray(mask)
                if mask.min() == 0 and par != "land_binary_mask":
                    first_mask_point = np.where(mask.ravel() == 0)[0][0]
                    if variables[par].ndim == 3:
                        upper = variables[par][0, :, :]
                    else:
                        upper = variables[par]
                    mask_values[par] = upper.ravel()[first_mask_point]
                    variables[par][variables[par] == mask_values[par]] = np.nan

            if var.ndim == 4:
                # Regrid from sigma to z levels
                if len(np.atleast_1d(indz)) > 1:
                    logger.debug("sigma to z for " + varname[0])
                    if self.precalculate_s2z_coefficients is True:
                        M = self.sea_floor_depth_below_sea_level.shape[0]
                        N = self.sea_floor_depth_below_sea_level.shape[1]
                        O = len(self.z_rho_tot)
                        if not hasattr(self, "s2z_A"):
                            logger.debug(
                                "Calculating sigma2z-coefficients for whole domain"
                            )
                            starttime = datetime.now()
                            dummyvar = np.ones((O, M, N))
                            dummy, self.s2z_total = depth.multi_zslice(
                                dummyvar, self.z_rho_tot, self.zlevels
                            )
                            # Store arrays/coefficients
                            self.s2z_A = self.s2z_total[0].reshape(
                                len(self.zlevels), M, N
                            )
                            self.s2z_C = self.s2z_total[1].reshape(
                                len(self.zlevels), M, N
                            )
                            # self.s2z_I = self.s2z_total[2].reshape(M, N)
                            self.s2z_kmax = self.s2z_total[3]
                            del self.s2z_total  # Free memory
                            logger.info("Time: " + str(datetime.now() - starttime))
                        if "A" not in locals():
                            logger.debug("Re-using sigma2z-coefficients")
                            # Select relevant subset of full arrays
                            zle = np.arange(zi1, zi2)  # The relevant depth levels
                            A = (
                                self.s2z_A.copy()
                            )  # Awkward subsetting to prevent losing one dimension
                            A = A[:, :, indx]
                            A = A[:, indy, :]
                            A = A[zle, :, :]
                            C = self.s2z_C.copy()
                            C = C[:, :, indx]
                            C = C[:, indy, :]
                            C = C[zle, :, :]
                            C = C - C.max() + variables[par].shape[0] - 1
                            C[C < 1] = 1
                            A = A.reshape(len(zle), len(indx) * len(indy))
                            C = C.reshape(len(zle), len(indx) * len(indy))
                            I = np.arange(len(indx) * len(indy))
                            ## Check
                            # dummyvar2, s2z = depth.multi_zslice(
                            #    variables[par].copy(), z_rho.copy(), variables['z'])
                            # print len(zle), variables[par].shape, 'zle, varshape'
                            # Ac,Cc,Ic,kmaxc = s2z
                            # print C, 'C'
                            # print Cc, 'Cc'
                            # print C.shape, Cc.shape
                            # if C.max() != Cc.max():
                            #    print 'WARNING!!'
                            #    import sys; sys.exit('stop')
                            kmax = len(
                                zle
                            )  # Must be checked. Or number of sigma-layers?
                    if "A" not in locals():
                        logger.debug("Calculating new sigma2z-coefficients")
                        variables[par], s2z = depth.multi_zslice(
                            variables[par], z_rho, variables["z"]
                        )
                        A, C, I, kmax = s2z
                        # Reshaping to compare with subset of full array
                        # zle = np.arange(zi1, zi2)
                        # A = A.reshape(len(zle), len(indx), len(indy))
                        # C = C.reshape(len(zle), len(indx), len(indy))
                        # I = I.reshape(len(indx), len(indy))
                    else:
                        logger.debug("Applying sigma2z-coefficients")
                        # Re-using sigma2z koefficients:
                        F = np.asarray(variables[par])
                        Fshape = F.shape
                        N = F.shape[0]
                        M = F.size // N
                        F = F.reshape((N, M))
                        R = (1 - A) * F[(C - 1, I)] + A * F[(C, I)]
                        variables[par] = R.reshape((kmax,) + Fshape[1:])

                    # Nan in input to multi_zslice gives extreme values in output
                    variables[par][variables[par] > 1e9] = np.nan

            # Mask values outside domain
            variables[par] = np.ma.array(variables[par], ndmin=2, mask=False)

            # Skipping de-staggering, as it leads to invalid values at later interpolation
            # if block is True:
            #    # Unstagger grid for vectors
            #    logger.debug('Unstaggering ' + par)
            #    if 'eta_v' in var.dimensions:
            #        variables[par] = np.ma.array(variables[par],
            #                            mask=variables[par].mask)
            #        variables[par][variables[par].mask] = 0
            #        if variables[par].ndim == 2:
            #            variables[par] = \
            #                (variables[par][0:-1,0:-1] +
            #                variables[par][0:-1,1::])/2
            #        elif variables[par].ndim == 3:
            #            variables[par] = \
            #                (variables[par][:,0:-1,0:-1] +
            #                variables[par][:,0:-1,1::])/2
            #        variables[par] = np.ma.masked_where(variables[par]==0,
            #                                            variables[par])
            #    elif 'eta_u' in var.dimensions:
            #        variables[par] = np.ma.array(variables[par],
            #                            mask=variables[par].mask)
            #        variables[par][variables[par].mask] = 0
            #        if variables[par].ndim == 2:
            #            variables[par] = \
            #                (variables[par][0:-1,0:-1] +
            #                 variables[par][1::,0:-1])/2
            #        elif variables[par].ndim == 3:
            #            variables[par] = \
            #                (variables[par][:,0:-1,0:-1] +
            #                 variables[par][:,1::,0:-1])/2
            #        variables[par] = np.ma.masked_where(variables[par]==0,
            #                                            variables[par])
            #    else:
            #        if variables[par].ndim == 2:
            #            variables[par] = variables[par][1::, 1::]
            #        elif variables[par].ndim == 3:
            #            variables[par] = variables[par][:,1::, 1::]

        # TODO: should be midpoints, but angle array below needs integer
        # indx = indx[0:-1]  # Only if de-staggering has been performed
        # indy = indy[1::]
        variables["x"] = indx
        variables["y"] = indy

        variables["x"] = variables["x"].astype(np.float32)
        variables["y"] = variables["y"].astype(np.float32)
        variables["time"] = nearestTime

        if (
            "x_sea_water_velocity" in variables.keys()
            or "sea_ice_x_velocity" in variables.keys()
            or "x_wind" in variables.keys()
            and "x_sea_water_velocity" not in self.do_not_rotate
            and "sea_ice_x_velocity" not in self.do_not_rotate
            and "x_wind" not in self.do_not_rotate
        ):
            # We must rotate current vectors
            if not hasattr(self, "angle_xi_east"):
                if "angle" in self.Dataset.variables:
                    logger.debug("Reading angle between xi and east...")
                    self.angle_xi_east = self.Dataset.variables["angle"][:]
            if isinstance(self.angle_xi_east, int):
                rad = self.angle_xi_east
            else:
                rad = self.angle_xi_east[indy, indx]
                rad = np.ma.asarray(rad)
            if (
                "x_sea_water_velocity" in variables.keys()
                and "x_sea_water_velocity" not in self.do_not_rotate
            ):
                (
                    variables["x_sea_water_velocity"],
                    variables["y_sea_water_velocity"],
                ) = rotate_vectors_angle(
                    variables["x_sea_water_velocity"],
                    variables["y_sea_water_velocity"],
                    rad,
                )
                logger.debug("Rotated x_sea_water_velocity and y_sea_water_velocity")
            if (
                "sea_ice_x_velocity" in variables.keys()
                and "sea_ice_x_velocity" not in self.do_not_rotate
            ):
                (
                    variables["sea_ice_x_velocity"],
                    variables["sea_ice_y_velocity"],
                ) = rotate_vectors_angle(
                    variables["sea_ice_x_velocity"],
                    variables["sea_ice_y_velocity"],
                    rad,
                )
                logger.debug("Rotated sea_ice_x_velocity and sea_ice_y_velocity")
            if "x_wind" in variables.keys() and "x_wind" not in self.do_not_rotate:
                variables["x_wind"], variables["y_wind"] = rotate_vectors_angle(
                    variables["x_wind"], variables["y_wind"], rad
                )
                logger.debug("Rotated x_wind and y_wind")

        # Masking NaN
        for var in requested_variables:
            variables[var] = np.ma.masked_invalid(variables[var])

        logger.debug("Time for ROMS native reader: " + str(datetime.now() - start_time))

        return variables


def rotate_vectors_angle(u, v, radians):
    """Rotate vectors to be east/north."""
    u2 = u * np.cos(radians) - v * np.sin(radians)
    v2 = u * np.sin(radians) + v * np.cos(radians)
    return u2, v2



"""Test realistic scenarios, which are slower."""

import pytest

import particle_tracking_manager as ptm


@pytest.mark.slow
def test_add_new_reader():
    """Add a separate reader from the defaults."""

    import xroms

    manager = ptm.OpenDriftModel()

    url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
    reader_kwargs = dict(loc=url, kwargs_xarray={})
    manager.add_reader(**reader_kwargs)


@pytest.mark.slow
def test_run():
    """Set up and run."""

    import xroms

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1)
    manager = ptm.OpenDriftModel(**seeding_kwargs)

    url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
    reader_kwargs = dict(loc=url, kwargs_xarray={})
    manager.add_reader(**reader_kwargs)
    # can find reader at manager.o.env.readers['roms native']

    manager.start_time = manager.o.env.readers["roms native"].start_time
    manager.seed()
    manager.run()



from datetime import datetime
from unittest import mock

import numpy as np
import xarray as xr

import particle_tracking_manager as ptm

from particle_tracking_manager.models.opendrift.reader_ROMS_native import Reader


# @mock.patch("xarray.open_mfdataset")
@mock.patch("xarray.open_dataset")
def test_init(mock_dataset):
    """Test initialization of ROMS reader."""

    ds = xr.Dataset()
    ds["ocean_time"] = (
        ("ocean_time"),
        [1.258632e09, 1.258646e09, 1.258660e09],
        {"units": "seconds since 1970-01-01"},
    )
    ds["lat_rho"] = (("eta_rho", "xi_rho"), np.array([[54, 55, 56], [55, 56, 57]]))
    ds["lon_rho"] = (
        ("eta_rho", "xi_rho"),
        np.array([[-151, -151, -151], [-150, -150, -150]]),
    )
    ds["salt"] = (("eta_rho", "xi_rho"), np.ones((2, 3)))
    ds["temp"] = (("eta_rho", "xi_rho"), np.ones((2, 3)))

    mock_dataset.return_value = ds

    reader = Reader(filename="test")


# @pytest.mark.slow
# def test_add_new_reader():
#     manager = ptm.OpenDriftModel()

#     url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
#     reader_kwargs = dict(loc=url, kwargs_xarray={})
#     manager.add_reader(**reader_kwargs)

#     assert len(manager.o.env.readers) == 1
#     assert "roms native" in manager.o.env.readers


# @pytest.mark.slow
# def test_get_variables():
#     manager = ptm.OpenDriftModel()

#     url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
#     reader_kwargs = dict(loc=url, kwargs_xarray={})
#     manager.add_reader(**reader_kwargs)

#     # Set up test data
#     requested_variables = ["sea_floor_depth_below_sea_level", "x_sea_water_velocity"]
#     time = datetime.now()
#     x = 10.0
#     y = 20.0
#     z = None

#     variables = manager.o.env.readers["roms native"].get_variables(
#         requested_variables, time, x, y, z
#     )

#     # Assert the returned variables
#     assert "sea_floor_depth_below_sea_level" in variables
#     assert "x_sea_water_velocity" in variables

#     # Assert the shape of the variables
#     assert variables["sea_floor_depth_below_sea_level"].shape == (1,)
#     assert variables["x_sea_water_velocity"].shape == (1,)

#     # Assert the values of the variables
#     assert variables["sea_floor_depth_below_sea_level"][0] == pytest.approx(10.0, abs=1e-6)
#     assert variables["x_sea_water_velocity"][0] == pytest.approx(20.0, abs=1e-6)



"""Test manager use in library, the default approach."""

from datetime import datetime, timedelta
from unittest import mock

import numpy as np
import pytest

import particle_tracking_manager as ptm


def test_order():
    """Have to configure before seeding."""

    with pytest.raises(KeyError):
        manager = ptm.OpenDriftModel()
        manager.run()


def test_seed():
    """make sure seeding works with no ocean model

    also compare two approaches for inputting info.
    """

    manager = ptm.OpenDriftModel(use_auto_landmask=True, number=1)
    manager.lon = -151
    manager.lat = 59
    manager.start_time = datetime(2000, 1, 1)

    # with pytest.raises(ValueError):
    #     manager.seed()

    manager.ocean_model = "test"
    manager.has_added_reader = True  # cheat to run test
    manager.seed()
    # look at elements with manager.o.elements_scheduled

    seeding_kwargs = dict(lon=-151, lat=59, start_time=datetime(2000, 1, 1))
    manager2 = ptm.OpenDriftModel(
        use_auto_landmask=True, number=1, ocean_model="test", **seeding_kwargs
    )
    manager2.has_added_reader = True  # cheat to run test
    manager2.seed()

    assert (
        manager.o.elements_scheduled.__dict__ == manager2.o.elements_scheduled.__dict__
    )


@mock.patch(
    "particle_tracking_manager.models.opendrift.model_opendrift.OpenDriftModel.reader_metadata"
)
def test_lon_check(mock_reader_metadata):
    """Test longitude check that is run when variable and reader are set."""

    # Check that longitude is checked as being within (mocked) reader values
    mock_reader_metadata.return_value = np.array([-150, -140, -130])

    m = ptm.OpenDriftModel(lon=0, lat=0)

    # this causes the check
    with pytest.raises(AssertionError):
        m.has_added_reader = True


@mock.patch(
    "particle_tracking_manager.models.opendrift.model_opendrift.OpenDriftModel.reader_metadata"
)
def test_start_time_check(mock_reader_metadata):
    """Test start_time check that is run when variable and reader are set."""

    # Check that start_time is checked as being within (mocked) reader values
    mock_reader_metadata.return_value = datetime(2000, 1, 1)

    m = ptm.OpenDriftModel(start_time=datetime(1999, 1, 1))

    # this causes the check
    with pytest.raises(AssertionError):
        m.has_added_reader = True


@mock.patch(
    "particle_tracking_manager.models.opendrift.model_opendrift.OpenDriftModel.reader_metadata"
)
def test_ocean_model_not_None(mock_reader_metadata):
    """Test that ocean_model can't be None."""

    # Use this to get through steps necessary for the test
    mock_reader_metadata.return_value = datetime(2000, 1, 1)

    m = ptm.OpenDriftModel()
    with pytest.raises(AssertionError):
        m.has_added_reader = True


def test_parameter_passing():
    """make sure parameters passed into package make it to simulation runtime."""

    ts = 2 * 3600
    diffmodel = "windspeed_Sundby1983"
    use_auto_landmask = True

    seed_kws = dict(
        lon=4.0,
        lat=60.0,
        radius=5000,
        number=100,
        start_time=datetime(2015, 9, 22, 6, 0, 0),
    )
    m = ptm.OpenDriftModel(
        use_auto_landmask=use_auto_landmask,
        time_step=ts,
        duration=timedelta(hours=10),
        diffusivitymodel=diffmodel,
        **seed_kws
    )

    # idealized simulation, provide a fake current
    m.o.set_config("environment:fallback:y_sea_water_velocity", 1)

    # seed
    m.seed()

    # run simulation
    m.run()

    # check time_step across access points
    assert (
        m.o.time_step.seconds
        == ts
        == m.time_step
        == m.show_config_model(key="time_step")["value"]
    )

    # check diff model
    assert m.show_config(key="diffusivitymodel")["value"] == diffmodel

    # check use_auto_landmask coming through
    assert m.show_config(key="use_auto_landmask")["value"] == use_auto_landmask


def test_keyword_parameters():
    """Make sure unknown parameters are not input."""

    with pytest.raises(KeyError):
        m = ptm.OpenDriftModel(incorrect_key="test")



"""Particle Tracking Manager."""

import cmocean

from .models.opendrift.model_opendrift import OpenDriftModel
from .the_manager import ParticleTrackingManager


cmap = cmocean.tools.crop_by_percent(cmocean.cm.amp, 20, which="max", N=None)



"""From Copilot"""

import unittest

from particle_tracking_manager.models.opendrift.model_opendrift import (
    OpenDriftModel,
    config_model,
)


class TestOpenDriftModel(unittest.TestCase):
    def setUp(self):
        self.odm = OpenDriftModel()

    def test_init(self):
        self.assertEqual(self.odm.drift_model, config_model["drift_model"]["default"])
        self.assertEqual(self.odm.radius, config_model["radius"]["default"])
        self.assertEqual(self.odm.radius_type, config_model["radius_type"]["default"])
        # self.assertEqual(self.odm.horizontal_diffusivity, config_model["horizontal_diffusivity"]["default"])
        self.assertEqual(
            self.odm.use_auto_landmask, config_model["use_auto_landmask"]["default"]
        )
        self.assertEqual(
            self.odm.diffusivitymodel, config_model["diffusivitymodel"]["default"]
        )
        self.assertEqual(self.odm.stokes_drift, config_model["stokes_drift"]["default"])
        self.assertEqual(
            self.odm.mixed_layer_depth, config_model["mixed_layer_depth"]["default"]
        )
        self.assertEqual(
            self.odm.coastline_action, config_model["coastline_action"]["default"]
        )
        self.assertEqual(self.odm.max_speed, config_model["max_speed"]["default"])
        self.assertEqual(
            self.odm.wind_drift_factor, config_model["wind_drift_factor"]["default"]
        )
        self.assertEqual(
            self.odm.wind_drift_depth, config_model["wind_drift_depth"]["default"]
        )
        self.assertEqual(
            self.odm.vertical_mixing_timestep,
            config_model["vertical_mixing_timestep"]["default"],
        )
        self.assertEqual(self.odm.object_type, config_model["object_type"]["default"])
        self.assertEqual(self.odm.diameter, config_model["diameter"]["default"])
        self.assertEqual(
            self.odm.neutral_buoyancy_salinity,
            config_model["neutral_buoyancy_salinity"]["default"],
        )
        self.assertEqual(
            self.odm.stage_fraction, config_model["stage_fraction"]["default"]
        )
        self.assertEqual(self.odm.hatched, config_model["hatched"]["default"])
        self.assertEqual(self.odm.length, config_model["length"]["default"])
        self.assertEqual(self.odm.weight, config_model["weight"]["default"])
        self.assertEqual(self.odm.oil_type, config_model["oil_type"]["default"])
        self.assertEqual(self.odm.m3_per_hour, config_model["m3_per_hour"]["default"])
        self.assertEqual(
            self.odm.oil_film_thickness, config_model["oil_film_thickness"]["default"]
        )
        self.assertEqual(
            self.odm.droplet_size_distribution,
            config_model["droplet_size_distribution"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_mu, config_model["droplet_diameter_mu"]["default"]
        )
        self.assertEqual(
            self.odm.droplet_diameter_sigma,
            config_model["droplet_diameter_sigma"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_min_subsea,
            config_model["droplet_diameter_min_subsea"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_max_subsea,
            config_model["droplet_diameter_max_subsea"]["default"],
        )
        self.assertEqual(
            self.odm.emulsification, config_model["emulsification"]["default"]
        )
        self.assertEqual(self.odm.dispersion, config_model["dispersion"]["default"])
        self.assertEqual(self.odm.evaporation, config_model["evaporation"]["default"])
        self.assertEqual(
            self.odm.update_oilfilm_thickness,
            config_model["update_oilfilm_thickness"]["default"],
        )
        self.assertEqual(
            self.odm.biodegradation, config_model["biodegradation"]["default"]
        )


if __name__ == "__main__":
    unittest.main()



"""Contains logic for configuring particle tracking simulations."""


# from docstring_inheritance import NumpyDocstringInheritanceMeta
import datetime
import logging
import pathlib
import warnings

from pathlib import Path
from typing import Optional, Union

import cmocean.cm as cmo
import numpy as np
import pandas as pd
import xarray as xr
import yaml

from .cli import is_None


# Read PTM configuration information
loc = pathlib.Path(__file__).parent / pathlib.Path("the_manager_config.yaml")
with open(loc, "r") as f:
    # Load the YAML file into a Python object
    config_ptm = yaml.safe_load(f)

# convert "None"s to Nones
for key in config_ptm.keys():
    if is_None(config_ptm[key]["default"]):
        config_ptm[key]["default"] = None


ciofs_operational_start_time = datetime.datetime(2021, 8, 31, 19, 0, 0)
ciofs_operational_end_time = (pd.Timestamp.now() + pd.Timedelta("48H")).to_pydatetime()
ciofs_end_time = datetime.datetime(2023, 1, 1, 0, 0, 0)
nwgoa_end_time = datetime.datetime(2009, 1, 1, 0, 0, 0)
overall_start_time = datetime.datetime(1999, 1, 1, 0, 0, 0)
overall_end_time = ciofs_operational_end_time


class ParticleTrackingManager:
    """Manager class that controls particle tracking model."""

    ocean_model: str
    lon: Union[int, float]
    lat: Union[int, float]
    surface_only: Optional[bool]
    z: Optional[Union[int, float]]
    start_time: Optional[datetime.datetime]
    log: str

    def __init__(
        self,
        model: str,
        lon: Optional[Union[int, float]] = None,
        lat: Optional[Union[int, float]] = None,
        z: Optional[Union[int, float]] = config_ptm["z"]["default"],
        seed_seafloor: bool = config_ptm["seed_seafloor"]["default"],
        number: int = config_ptm["number"]["default"],
        start_time: Optional[datetime.datetime] = None,
        run_forward: bool = config_ptm["run_forward"]["default"],
        time_step: int = config_ptm["time_step"]["default"],
        time_step_output: Optional[int] = config_ptm["time_step_output"]["default"],
        steps: Optional[int] = config_ptm["steps"]["default"],
        duration: Optional[datetime.timedelta] = config_ptm["duration"]["default"],
        end_time: Optional[datetime.datetime] = config_ptm["end_time"]["default"],
        # universal inputs
        log: str = config_ptm["log"]["default"],
        ocean_model: Optional[str] = config_ptm["ocean_model"]["default"],
        surface_only: Optional[bool] = config_ptm["surface_only"]["default"],
        do3D: bool = config_ptm["do3D"]["default"],
        vertical_mixing: bool = config_ptm["vertical_mixing"]["default"],
        **kw,
    ) -> None:
        """Inputs necessary for any particle tracking.

        Parameters
        ----------
        model : str
            Name of Lagrangian model package to use for drifter tracking. Only option
            currently is "opendrift".
        lon : Optional[Union[int,float]], optional
            Longitude of center of initial drifter locations, by default None
        lat : Optional[Union[int,float]], optional
            Latitude of center of initial drifter locations, by default None
        z : Union[int,float], optional
            Depth of initial drifter locations, by default 0 but taken from the
            default in the model. Values are overridden if
            ``surface_only==True`` to 0 and to the seabed if ``seed_seafloor`` is True.
        seed_seafloor : bool, optional
            Set to True to seed drifters vertically at the seabed, default is False. If True
            then value of z is set to None and ignored.
        number : int
            Number of drifters to simulate. Default is 100.
        start_time : Optional[datetime], optional
            Start time of simulation, as a datetime object, by default None
        run_forward : bool, optional
            True to run forward in time, False to run backward, by default True
        time_step : int, optional
            Time step in seconds, options >0, <86400 (1 day in seconds), by default 3600
        time_step_output : int, optional
            How often to output model output, in seconds. Should be a multiple of time_step.
            By default will take the value of time_step (this change occurs in the model).
        steps : int, optional
            Number of time steps to run in simulation. Options >0.
            steps, end_time, or duration must be input by user. By default steps is 3 and
            duration and end_time are None.
        duration : Optional[datetime.timedelta], optional
            Length of simulation to run, as positive-valued timedelta object, in hours,
            such as ``timedelta(hours=48)``.
            steps, end_time, or duration must be input by user. By default steps is 3 and
            duration and end_time are None.
        end_time : Optional[datetime], optional
            Datetime at which to end simulation, as positive-valued datetime object.
            steps, end_time, or duration must be input by user. By default steps is 3 and
            duration and end_time are None.
        log : str, optional
            Options are "low" and "high" verbosity for log, by default "low"
        ocean_model : Optional[str], optional
            Name of ocean model to use for driving drifter simulation, by default None.
            Use None for testing and set up. Otherwise input a string.
            Options are: "NWGOA", "CIOFS", "CIOFS_now".
            Alternatively keep as None and set up a separate reader (see example in docs).
        surface_only : bool, optional
            Set to True to keep drifters at the surface, by default None.
            If this flag is set to not-None, it overrides do3D to False, vertical_mixing to False,
            and the z value(s) 0.
            If True, this flag also turns off reading model output below 0.5m if
            drift_model is not Leeway:
            ``o.set_config('drift:truncate_ocean_model_below_m', 0.5)`` to save time.
        do3D : bool, optional
            Set to True to run drifters in 3D, by default False. This is overridden if
            ``surface_only==True``. If True, vertical advection and mixing are turned on with
            options for setting ``diffusivitymodel``, ``background_diffusivity``,
            ``ocean_mixed_layer_thickness``, ``vertical_mixing_timestep``. If False,
            vertical motion is disabled.
        vertical_mixing : bool, optional
            Set to True to include vertical mixing, by default False. This is overridden if
            ``surface_only==True``.

        Notes
        -----
        Configuration happens at initialization time for the child model. There is currently
        no separate configuration step.
        """

        # get all named parameters input to ParticleTrackingManager class
        from inspect import signature

        sig = signature(ParticleTrackingManager)

        self.config_ptm = config_ptm

        self.logger = logging.getLogger(model)

        # Set all attributes which will trigger some checks and changes in __setattr__
        # these will also update "value" in the config dict
        for key in sig.parameters.keys():
            self.__setattr__(key, locals()[key])

        # mode flags
        self.has_added_reader = False
        self.has_run_seeding = False
        self.has_run = False

        self.kw = kw

    def __setattr_model__(self, name: str, value) -> None:
        """Implement this in model class to add specific __setattr__ there too."""
        pass

    def __setattr__(self, name: str, value) -> None:
        """Implement my own __setattr__ to enforce subsequent actions."""

        # create/update class attribute
        self.__dict__[name] = value

        # create/update "value" keyword in config to keep it up to date
        if name != "config_ptm" and name in self.config_ptm.keys():
            self.config_ptm[name]["value"] = value

        # create/update "value" keyword in model config to keep it up to date
        if hasattr(self, "config_model"):  # can't run this until init in model class
            self.__setattr_model__(name, value)

        # check longitude when it is set
        if value is not None and name == "lon":
            assert (
                -180 <= value <= 180
            ), "Longitude needs to be between -180 and 180 degrees."

        if value is not None and name == "lat":
            assert (
                -90 <= value <= 90
            ), "Latitude needs to be between -90 and 90 degrees."

        # make sure ocean_model name uppercase
        if name == "ocean_model" and value is not None:
            self.__dict__[name] = value.upper()
            self.config_ptm["ocean_model"]["value"] = value.upper()

        # check start_time relative to ocean_model selection
        if name in ["ocean_model", "start_time"]:
            if (
                hasattr(self, "start_time")
                and self.start_time is not None
                and hasattr(self, "ocean_model")
                and self.ocean_model is not None
            ):
                if value == "NWGOA":
                    assert overall_start_time <= value <= nwgoa_end_time
                elif value == "CIOFS":
                    assert overall_start_time <= value <= ciofs_end_time
                elif value == "CIOFS_NOW":
                    assert (
                        ciofs_operational_start_time
                        <= value
                        <= ciofs_operational_end_time
                    )

        # deal with if input longitudes need to be shifted due to model
        if name == "oceanmodel_lon0_360" and value:
            if self.ocean_model is not "test" and self.lon is not None:
                # move longitude to be 0 to 360 for this model
                # this is not a user-defined option
                if -180 < self.lon < 0:
                    self.__dict__["lon"] += 360

        if name == "surface_only" and value:
            self.logger.info(
                "overriding values for `do3D`, `z`, and `vertical_mixing` because `surface_only` True"
            )
            self.do3D = False
            self.z = 0
            self.vertical_mixing = False

        # in case any of these are reset by user after surface_only is already set
        if name in ["do3D", "z", "vertical_mixing"]:
            if hasattr(self, "surface_only") and self.surface_only:
                self.logger.info(
                    "overriding values for `do3D`, `z`, and `vertical_mixing` because `surface_only` True"
                )
                if name == "do3D":
                    value is False
                if name == "z":
                    value is 0
                if name == "vertical_mixing":
                    value is False
                self.__dict__[name] = value
                self.config_ptm[name]["value"] = value

            # if not 3D turn off vertical_mixing
            if hasattr(self, "do3D") and not self.do3D:
                self.logger.info("turning off vertical_mixing since do3D is False")
                self.__dict__["vertical_mixing"] = False
                self.config_ptm["vertical_mixing"]["value"] = False
                # self.vertical_mixing = False  # this is recursive

        # set z to None if seed_seafloor is True
        if name == "seed_seafloor" and value:
            self.logger.info("setting z to None since being seeded at seafloor")
            self.z = None

        # in case z is changed back after initialization
        if name == "z" and value is not None and hasattr(self, "seed_seafloor"):
            self.logger.info(
                "setting `seed_seafloor` to False since now setting a non-None z value"
            )
            self.seed_seafloor = False

        # if reader, lon, and lat set, check inputs
        if (
            name == "has_added_reader"
            and value
            and self.lon is not None
            and self.lat is not None
            or name in ["lon", "lat"]
            and hasattr(self, "has_added_reader")
            and self.has_added_reader
            and self.lon is not None
            and self.lat is not None
        ):

            if self.ocean_model != "TEST":
                rlon = self.reader_metadata("lon")
                assert rlon.min() < self.lon < rlon.max()
                rlat = self.reader_metadata("lat")
                assert rlat.min() < self.lat < rlat.max()

        # use reader start time if not otherwise input
        if name == "has_added_reader" and value and self.start_time is None:
            self.logger.info("setting reader start_time as simulation start_time")
            self.start_time = self.reader_metadata("start_time")

        # if reader, lon, and lat set, check inputs
        if name == "has_added_reader" and value and self.start_time is not None:

            if self.ocean_model != "TEST":
                assert self.reader_metadata("start_time") <= self.start_time

        # if reader, lon, and lat set, check inputs
        if name == "has_added_reader" and value:
            assert self.ocean_model is not None

    def add_reader(self, **kwargs):
        """Here is where the model output is opened."""

        self.run_add_reader(**kwargs)

        self.has_added_reader = True

    def seed(self, lon=None, lat=None, z=None):
        """Initialize the drifters in space and time

        ... and with any special properties.
        """

        for key in [lon, lat, z]:
            if key is not None:
                self.__setattr__(self, f"{key}", key)

        # if self.ocean_model != "TEST" and not self.has_added_reader:
        #     raise ValueError("first add reader with `manager.add_reader(**kwargs)`.")

        msg = f"""lon and lat need non-None values.
                  Update them with e.g. `self.lon=-151` or input to `seed`."""
        assert self.lon is not None and self.lat is not None

        msg = f"""z needs a non-None value.
                  Please update it with e.g. `self.z=-10` or input to `seed`."""
        if not self.seed_seafloor:
            assert self.z is not None, msg

        self.run_seed()
        self.has_run_seeding = True

    def run(self):
        """Call model run function."""

        if not self.has_run_seeding:
            raise KeyError("first run seeding with `manager.seed()`.")

        # need end time info
        assert (
            self.steps is not None
            or self.duration is not None
            or self.end_time is not None
        )

        if self.run_forward:
            timedir = 1
        else:
            timedir = -1

        if self.steps is not None:
            self.end_time = self.start_time + timedir * self.steps * datetime.timedelta(
                seconds=self.time_step
            )
            self.duration = abs(self.end_time - self.start_time)
        elif self.duration is not None:
            self.end_time = self.start_time + timedir * self.duration
            self.steps = self.duration / datetime.timedelta(seconds=self.time_step)
        elif self.end_time is not None:
            self.duration = abs(self.end_time - self.start_time)
            self.steps = self.duration / datetime.timedelta(seconds=self.time_step)

        self.run_drifters()
        self.has_run = True

    def run_all(self):
        """Run all steps."""

        if not self.has_added_reader:
            self.add_reader()

        if not self.has_run_seeding:
            self.seed()

        if not self.has_run:
            self.run()

    def output(self):
        """Hold for future output function."""
        pass

    def _config(self):
        """Model should have its own version which returns variable config"""
        pass

    def _add_ptm_config(self):
        """Have this in the model class to modify config"""
        pass

    def _add_model_config(self):
        """Have this in the model class to modify config"""
        pass

    def _update_config(self) -> None:
        """Update configuration between model, PTM additions, and model additions."""

        # Modify config with PTM config
        self._add_ptm_config()

        # Modify config with model-specific config
        self._add_model_config()

    def show_config_model(self):
        """define in child class"""
        pass

    def show_config(self, **kwargs) -> dict:
        """Show parameter configuration across both model and PTM."""

        self._update_config()

        # Filter config
        config = self.show_config_model(**kwargs)

        return config

    def reader_metadata(self, key):
        """define in child class"""
        pass

    def query_reader(self):
        """Overwrite method in model."""
        pass
